/*
** Module   : ServerImpl.java
** Abstract : A class that implements the app server specific method and  attribute APIs.
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CS  20130211 Created initial version.
** 002 CA  20130221 Added resourceDelete stub. setType was renamed to setResourceType. 
** 003 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CS  20130321 Added missing versions of connect() method.
** 005 CA  20130529 Added runtime support for appservers.
** 006 CA  20130822 Added support for async requests.
** 007 CA  20130908 Moved connect-related code to ConnectHelper.
** 008 CA  20130927 Resource type is computed from the annotation. Removed setUnknown(false) call.
** 009 OM  20131018 Updated to conform the new SessionListener API.
** 010 CA  20131220 Refactored to hide the type of the connection (appserver or web service) in a
**                  ServerHelper implementation. Added support for the web service connection.
** 011 CA  20140326 Fixed a NPE bug when cleaning up async requests for a session-managed 
**                  connection.
** 012 VIG 20141213 Added SUBTYPE setters stubs.
*/

package com.goldencode.p2j.util;

import java.net.*;
import java.util.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.security.*;

/**
 * A class for implementing the Server methods and attributes.
 */
public class ServerImpl
extends HandleChain 
implements Server,
           AsyncRequestCountAttribute,
           FirstLastProcedureAttribute,
           SslServerAttribute,
           SubTypeAttribute,
           ConnectableServer,
           SessionListener,
           AsyncRequestListener
{
   /**
    * The server helper for this server, set after the {@link #connect} method is called.
    */
   private ServerHelper helper = null;
   
   /** The first procedure in this server's proxy procedure list. */
   private ProxyProcedureWrapper firstProcedure = null;
   
   /** The last procedure in this server's proxy procedure list. */
   private ProxyProcedureWrapper lastProcedure = null;

   /** Flag indicating if the server is valid. */
   private boolean valid = true;
   
   /** Count the number of active async requests. */
   private long asyncRequestCount = 0;
   
   /**
    * The currently active timer thread which will cancel the async requests.
    * May be <code>null</code> if no timer is active.
    */
   private Thread cancelReqTimer = null;
   
   /** An special instance used to perform async request. */
   private AsyncInvoker invoker = null;
   
   /**
    * Default constructor which initializes Server specific settings.
    */
   public ServerImpl()
   {
      // nothing special to do
   }
   
   /**
    * Check if the given handle refers a server resource.
    * 
    * @param    h
    *
    * @return   <code>true</code> if the handle's resource is a {@link ServerImpl server}.  
    */
   public static boolean isServer(handle h)
   {
      return h != null && !h.isUnknown() && (h.get() instanceof ServerImpl);
   }

   /**
    * Invoke the given request on this server.
    * 
    * @param    request
    *           The async request to be executed.
    */
   public void invokeAsync(AsyncRequestImpl request)
   {
      AsyncInvoker invoker = getAsyncInvoker();
      invoker.invoke(request);
   }
   
   /**
    * Reports if this object is valid for use (has not been deleted).
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return valid;
   }

   /**
    * Returns the number of asynchronous requests.
    * 
    * @return   See above.
    */
   public synchronized integer getAsyncRequestCount()
   {
      return new integer(asyncRequestCount);
   }

   /**
    * Returns the subtype of the handle. 
    * <p>
    * For SERVER handles this represents the type of server to which this is connected. 
    * The possible values are "APPSERVER" and "WEBSERVICE".
    * 
    * @return   See above.
    */
   public character getSubType()
   {
      return new character(helper.getSubType());
   }

   /**
    * Sets the the subtype of the object.
    * 
    * @param    value
    *           The new value of the subtype object attribute.
    */
   public void setSubType(character value)
   {
      handle.readOnlyError(new handle(this), "SUB-TYPE");
   }
   
   /**
    * Sets the the subtype of the object.
    * 
    * @param    value
    *           The new value of the subtype object attribute.
    */
   public void setSubType(String value)
   {
      handle.readOnlyError(new handle(this), "SUB-TYPE");
   }
   
   /**
    * Java implementation for getting the  SSL-SERVER-NAME  attribute  which contains the SSL 
    * server name if the connection is made as SSL.
    * 
    * @return   The SSL server name, if the connection is made as SSL or unknown
    *           otherwise
    */
   public character getSslServerName()
   {
      // TODO Implement the actual logic.
      return new character();
   }

   /**
    * Returns the first entry in the list of persistent procedures or procedure objects.
    * 
    * @return   See above.
    */
   public handle firstProcedure()
   {
      return new handle(firstProcedure);
   }

   /**
    * Returns the first entry in the list of persistent procedures or procedure objects.
    * 
    * @return   See above.
    */
   public handle lastProcedure()
   {
      return new handle(lastProcedure);
   }

   /**
    * This returns the id for the AppServer connection to this handle. For Web Services, the
    * returned  value will be an empty string.
    * 
    * @return   See above.
    */
   public character getClientConnectionId()
   {
      return new character(!_connected() ? "" : helper.getConnectionID());
   }

   /**
    * Returns a handle to the first asynchronous requests in the AppServer or WebService list of
    * asynchronous requests.
    * 
    * @return   See above.
    */
   public handle getFirstAsyncRequest()
   {
      return new handle(HandleChain.firstResource(LegacyResource.ASYNC_REQUEST));
   }

   /**
    * Returns a handle to the last asynchronous requests in the AppServer or WebService list of
    * asynchronous requests.
    * 
    * @return   See above.
    */
   public handle getLastAsyncRequest()
   {
      return new handle(HandleChain.lastResource(LegacyResource.ASYNC_REQUEST));
   }
   
   /**
    * Cancel all running or queued requests.  Depending on the server's operating mode, the active
    * requests will treat a cancel command differently:
    * <ul>
    *    <li>Session-managed: STOP is sent to the executing request; the COMPLETE flag will remain
    *        set to TRUE. All the queued requests are purged and their CANCELLED flag is set to
    *        TRUE.</li>
    *    <li>Session-free: STOP is sent to all requests; their CANCELLED flag is set to TRUE.</li> 
    * </ul>
    * 
    * @return   <code>true</code> if CANCEL-REQUESTS() method returns <code>true</code>, 
    *           <code>false</code> otherwise.
    */
   public logical cancelRequests()
   {
      if (!_connected())
      {
         return new logical(false);
      }
      
      AsyncInvoker invoker = getAsyncInvoker();
      return new logical(invoker.cancelRequests());
   }

   /**
    * Calls the {@link #cancelRequests()} method when the specified number of seconds have passed.
    * 
    * @param    timeSec 
    *           The time in seconds after the {@link #cancelRequests()} method will be called. 4GL
    *           allows only integer or int64 values for this parameter. 
    *         
    * @return   <code>false</code> if timer cannot be set or {@link #cancelRequests()}'s result 
    *           otherwise
    */
   public logical cancelRequestsAfter(integer timeSec)
   {
      return cancelRequestsAfter(timeSec.longValue());
   }
   
   /**
    * Calls the {@link #cancelRequests()} method when the specified number of seconds have passed.
    * 
    * @param    timeSec 
    *           The time in seconds after the {@link #cancelRequests()} method will be called. 4GL
    *           allows only integer or int64 values for this parameter. 
    *         
    * @return   <code>false</code> if timer cannot be set or {@link #cancelRequests()}'s result 
    *           otherwise
    */
   public logical cancelRequestsAfter(int64 timeSec)
   {
      return cancelRequestsAfter(timeSec.longValue());
   }
   
   /**
    * Calls the {@link #cancelRequests()} method when the specified number of seconds have passed;
    * if another timer exists for this server, it will be terminated and a new timer will start.
    * <p>
    * If <code>timeSec</code> is zero or negative, then the current {@link #cancelReqTimer timer}
    * will be stopped. 
    * 
    * @param    timeSec 
    *           The time in seconds after the {@link #cancelRequests()} method will be called. 4GL
    *           allows only integer or int64 values for this parameter. 
    *         
    * @return   <code>false</code> if timer cannot be set or {@link #cancelRequests()}'s result 
    *           otherwise
    */
   public synchronized logical cancelRequestsAfter(long timeSec)
   {
      if (!_connected())
      {
         return new logical(false);
      }

      final long millis = timeSec * 1000;
      if (millis <= 0 && cancelReqTimer != null)
      {
         cancelReqTimer.interrupt();
         return new logical(true);
      }

      Runnable task = new Runnable()
      {
         @Override
         public void run()
         {
            try
            {
               Thread.sleep(millis);
               
               // time has elapsed, send cancel to all currently running or scheduled requests
               cancelRequests();
            }
            catch (InterruptedException e)
            {
               // stopped, nothing to do
            }
            
            synchronized (ServerImpl.this)
            {
               // clear the timer
               cancelReqTimer = null;
            }
         }
      };
      
      // if there is an existing timer, stop it
      if (cancelReqTimer != null)
      {
         cancelReqTimer.interrupt();
      }

      // start the new timer
      cancelReqTimer = new AssociatedThread(task);
      String serverName = helper.getServerName();
      String timerName = String.format("Cancel requests timer [%s]", serverName);
      String name = LogHelper.createThreadName(timerName);
      cancelReqTimer.setName(name);
      cancelReqTimer.start();
      
      return new logical(true);
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only 
    *           {@link character} values for this parameter.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise
    */
   public logical connect(String options)
   {
      return connect(new character(options));
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only
    *           character values for this parameter.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise
    */
   public logical connect(character options)
   {
      return connect(options, new character(), new character(), new character());
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only  
    *           character values for this parameter.
    * @param    user
    *           The user to send to the CONNECT procedure on the appserver side.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise
    */
   public logical connect(character options, character user)
   {
      return connect(options, user, new character(), new character());
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only  
    *           character values for this parameter.
    * @param    user
    *           The user to send to the CONNECT procedure on the appserver side.
    * @param    pwd
    *           The password to send to the CONNECT procedure on the appserver side.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code>
    *           otherwise
    */
   public logical connect(character options, character user, character pwd)
   {
      return connect(options, user, pwd, new character());
   }

   /**
    * Method declaration  for the CONNECT() procedure supported by web-services and server handles.
    *  
    * @param    options
    *           The set of parameters to establish the connection. 4GL allows only  
    *           character values for this parameter.
    * @param    user
    *           The user to send to the CONNECT procedure on the appserver side.
    * @param    pwd
    *           The password to send to the CONNECT procedure on the appserver side.
    * @param    serverInfo
    *           Other info to send to the CONNECT procedure on the appserver side.
    * 
    * @return   <code>true</code> if connection succeeded, or <code>false</code> otherwise
    */
   public logical connect(character options, character user, character pwd, character serverInfo)
   {
      if (_connected())
      {
         ErrorManager.displayError(5535, "Server is already connected", false);
         return new logical(false);
      }
      
      name.setUnknown();

      synchronized (this)
      {
         helper = null;
      }

      Map<String, String> parms = parseOptions(options);
      if (parms == null)
      {
         return new logical(false);
      }
      
      boolean res = false;
      // check the -WSDL parameter, which is required for web services
      boolean webService = parms.containsKey(WebServiceHelper.CONNECT_OPTION_WSDL);
      if (webService)
      {
         // the user, pwd and serverInfo params are ignored
         res = connectToWebService(parms);
      }
      else
      {
         res = connectToAppServer(parms, user, pwd, serverInfo);
      }
      
      if (!res)
      {
         return new logical(false);
      }
      
      // register this server with the session
      Session sess = SessionManager.get().getSession();
      sess.addSessionListener(this);
      helper.setServer(this);
      name.assign(helper.getServerName());

      return new logical(true);
   }
   
   /**
    * Method declaration for the CONNECTED() method, supported by web-services and server handles.
    * This will return <code>true</code> if the web service is connected or <code>false</code>
    * otherwise.
    * 
    * @return   <code>true</code> if the server is connected or <code>false</code> otherwise.
    */
   public logical connected()
   {
      return new logical(_connected());
   }

   /**
    * Method declaration for the CONNECTED() method, supported by web-services and server handles.
    * This will return <code>true</code> if the web service is connected or <code>false</code>
    * otherwise.
    * 
    * @return   <code>true</code> if the server  is connected or <code>false</code> otherwise.
    */
   public synchronized boolean _connected()
   {
      return helper != null;
   }

   /**
    * Method declaration for the DISCONNECT() method supported by web-services and server handles.
    * This will attempt to disconnect the resource and returns <code>true</code> if succeeded or 
    * <code>false</code> otherwise.
    * 
    * @return   <code>true</code> if disconnect succeeded or <code>false</code> otherwise.
    */
   public logical disconnect()
   {
      return disconnectImpl(false);
   }

   /**
    * This method is called when the session is ending.  It disconnects the current server, if
    * the session which created it is ending.
    *
    * @param    session
    *           The session that is ending.
    */
   @Override
   public void terminate(Session session)
   {
      if (_connected())
      {
         disconnectImpl(true);
      }
   }

   /**
    * This method is called when the session is starting, just after the context was created.
    * @param   session
    *          The session that is starting.
    */
   @Override
   public void initialize(Session session)
   {
      // no op
   }

   /**
    * Notification that a remote procedure is invoked via this server.
    */
   @Override
   public synchronized void notifyStart()
   {
      asyncRequestCount++;
   }

   /**
    * Notification that a remote procedure invoked via this server is finished.
    */
   @Override
   public synchronized void notifyFinish()
   {
      asyncRequestCount--;
   }
   
   /**
    * Check if the server has running async requests. If so, display appropriate error message.
    * 
    * @return   <code>true</code> if the server has running async requests.
    */
   synchronized boolean hasAsyncRequests()
   {
      if (asyncRequestCount != 0)
      {
         String msg = "Server %s has outstanding asynchronous requests.  Cannot RUN";
         msg = String.format(msg, helper.getConnectionID());
         ErrorManager.recordOrThrowError(9004, msg, false);
         
         return true;
      }
      
      return false;
   }

   /**
    * Get the {@link ServerHelper helper} for this server resource.
    * 
    * @return   See above.
    */
   synchronized ServerHelper getServerHelper()
   {
      return helper;
   }

   /**
    * Check if this server is connected to a web-service.
    * 
    * @return   See above.
    */
   boolean isWebService()
   {
      return helper.isWebService();
   }

   /**
    * Set the first proxy procedure of this server's chain.
    * 
    * @param    proc
    *           The procedure instance.
    */
   void setFirstProcedure(ProxyProcedureWrapper proc)
   {
      this.firstProcedure = proc;
   }
   
   /**
    * Set the last proxy procedure of this server's chain.
    * 
    * @param    proc
    *           The procedure instance.
    */
   void setLastProcedure(ProxyProcedureWrapper proc)
   {
      this.lastProcedure = proc;
   }
   
   /**
    * Delete the resource. If the server is still connected, delete is not possible.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   protected boolean resourceDelete()
   {
      if (_connected())
      {
         final String msg = "SERVER %s is still connected.  Cannot DELETE";
         final String err = String.format(msg, getClientConnectionId().toStringMessage());
         ErrorManager.recordOrThrowError(5457, err);

         return false;
      }

      this.valid = false;
      return true;
   }

   /**
    * Connect to a web service, using the specified parameters.
    * 
    * @param    parms
    *           The connection parameters.
    *           
    * @return   <code>true</code> if the connection was possible.
    */
   private boolean connectToWebService(Map<String, String> parms)
   {
      synchronized (this)
      {
         helper = WebServiceHelper.connect(parms);
      }

      return _connected();
   }

   /**
    * Connect to an appserver using the specified parameters.
    * 
    * @param    parms
    *           The set of parameters to establish the connection. 4GL allows only character
    *           values for this parameter.
    * @param    user
    *           The user to send to the CONNECT procedure on the appserver side.
    * @param    pwd
    *           The password to send to the CONNECT procedure on the appserver side.
    * @param    serverInfo
    *           Other info to send to the CONNECT procedure on the appserver side.
    *
    * @return   <code>true</code> if the connection was possible.
    */
   private boolean connectToAppServer(Map<String, String> parms,
                                      character           user,
                                      character           pwd,
                                      character           serverInfo)
   {
      String appService = parms.get("-AppService");
      String host = parms.get("-H");
      String serviceName = parms.get("-S");
      int port = 0;
      try
      {
         port = Integer.parseInt(serviceName);
      }
      catch (NumberFormatException e)
      {
         // ignore
      }
      String url = parms.get("-URL");
      if (url != null)
      {
         // get them from the URL
         String err = null;
         try
         {
            URL u = new URL(url);
            String protocol = u.getProtocol();
            if (!(protocol.equals("AppServer") ||
                  protocol.equals("http")      ||
                  protocol.equals("https")))
            {
               // if we are not connecting to an AppServer, than abort
               err = protocol;
            }
            else
            {
               host = u.getHost();
               port = u.getPort();
               serviceName = null;
            }
         }
         catch (MalformedURLException e)
         {
            err = url;
         }
         
         if (err != null)
         {
            ErrorManager.displayError(String.format("Invalid URL syntax: %s", err));
            ErrorManager.displayError(5468, "Application server connect failure", false);
            return false;
         }
      }
      boolean directConnect = parms.containsKey("-DirectConnect");
      
      String sessionModel = ConnectHelper.getString(parms, "-sessionModel", "Session-managed");
      boolean sessionFree = "Session-free".equalsIgnoreCase(sessionModel);
      // TODO: in session-free mode, is possible to establish more than one connection (to allow
      //       async requests). the following parameters are related to session-free mode 
      int connectionLifetime = ConnectHelper.getInt(parms, "-connectionLifetime", 300);
      int initialConnections = ConnectHelper.getInt(parms, "-initialConnections", 1);
      int maxConnections = ConnectHelper.getInt(parms, "-maxConnections", 0);
      // the -ns* options are not used by P2J

      try
      {
         synchronized (this)
         {
            this.helper = AppServerHelper.connectLegacyMode(appService, host, serviceName, port, 
                                                            directConnect, sessionFree,
                                                            user, pwd, serverInfo);
         }
      }
      catch (NumberedException e)
      {
         ErrorManager.recordOrThrowError(e);
      }
      finally
      {
         if (!_connected())
         {
            final String msg = "Application server connect failure";
            ErrorManager.displayError(5468, msg, false);
   
            return false;
         }
      }
      
      return true;
   }

   /**
    * Method declaration for the DISCONNECT() method supported by web-services and server handles.
    * This will attempt to disconnect the resource and returns <code>true</code> if succeeded or 
    * <code>false</code> otherwise.
    * 
    * @return   <code>true</code> if disconnect succeeded or <code>false</code> otherwise.
    */
   private logical disconnectImpl(boolean disconnecting)
   {
      if (!_connected())
      {
         ErrorManager.displayError(5537, "Server is not connected.", false);
         return new logical(false);
      }
      
      // if there is an existing timer, stop it
      if (cancelReqTimer != null)
      {
         cancelReqTimer.interrupt();
      }
      
      // cancel requests before deleting
      cancelRequests();

      // now delete all requests
      handle async = getFirstAsyncRequest();
      while (!async.isUnknown())
      {
         HandleOps.delete(async);
         async = getFirstAsyncRequest();
      }
      
      helper.disconnect();
      
      synchronized (this)
      {
         helper = null;
      }
   
      if (invoker != null)
      {
         invoker.terminate();
         invoker = null;
      }
      
      if (!disconnecting)
      {
         Session sess = SessionManager.get().getSession();
         sess.removeSessionListener(this);
      }
   
      return new logical(true);
   }

   /**
    * Parse and validate the given options; a map is returned with each parameters's value. If a
    * parameter doesn't have a value, the map will contain <code>null</code>.
    *  
    * @param    options
    *           The options from which the parameters will be extracted.
    *           
    * @return   The parameter-to-value map.
    */
   private Map<String, String> parseOptions(character options)
   {
      if (options.isUnknown())
      {
         // if unknown, it assumes it connects to an AppServer, and not WebService
         return new HashMap<String, String>();
      }
      
      final Set<String> knownAppServerOptions = new HashSet<String>();
      knownAppServerOptions.add("-AppService");
      knownAppServerOptions.add("-H");
      knownAppServerOptions.add("-S");
      knownAppServerOptions.add("-DirectConnect"); // no value
      knownAppServerOptions.add("-ssl"); // no value
      knownAppServerOptions.add("-nosessionreuse"); // no value
      knownAppServerOptions.add("-nohostverify"); // no value
      knownAppServerOptions.add("-pf");
      knownAppServerOptions.add("-AppServerKeepalive");
      knownAppServerOptions.add("-URL");
      knownAppServerOptions.add("-sessionModel");
      knownAppServerOptions.add("-connectionLifetime");
      knownAppServerOptions.add("-initialConnections");
      knownAppServerOptions.add("-maxConnections");
      knownAppServerOptions.add("-nsClientMaxPort");
      knownAppServerOptions.add("-nsClientMinPort");
      knownAppServerOptions.add("-nsClientPicklistExpiration");
      knownAppServerOptions.add("-nsClientPicklistSize");
      knownAppServerOptions.add("-nsClientPortRetry");
      knownAppServerOptions.add("-nsClientDelay");
      
      final Set<String> noValueOptions = new HashSet<String>();
      noValueOptions.add("-DirectConnect"); // no value
      noValueOptions.add("-ssl"); // no value
      noValueOptions.add("-nosessionreuse"); // no value
      noValueOptions.add("-nohostverify"); // no value
      
      final Set<String> knownWebServiceOptions = WebServiceHelper.getKnownConnectOptions();
      
      final Map<String, Integer> intParams = new HashMap<String, Integer>();
      intParams.put("-connectionLifetime", 0);
      intParams.put("-initialConnections", 0);
      intParams.put("-maxConnections", 0);
      
      String s = options.getValue();
      
      String[] opts = s.split(" ");
      // determine the target; assume appServer
      Set<String> knownOptions = knownAppServerOptions;
      for (int i = 0; i < opts.length; i++)
      {
         if (WebServiceHelper.CONNECT_OPTION_WSDL.equals(opts[i]))
         {
            knownOptions = knownWebServiceOptions;
            break;
         }
      }

      return ConnectHelper.parseOptions(options, knownOptions, noValueOptions, intParams);
   }

   /**
    * Get the {@link #invoker asynchronous invoker} for this server; if {@link #invoker} is null,
    * it will create a {@link SessionManagedInvoker} if the server is in the Session-managed
    * operating mode or a {@link SessionFreeInvoker} if the server is in the Session-free 
    * operating mode.
    * 
    * @return   See above.
    */
   private AsyncInvoker getAsyncInvoker()
   {
      if (invoker == null)
      {
         invoker = (helper.isSessionFree() ? new SessionFreeInvoker() 
                                           : new SessionManagedInvoker());
         
         // start in a different thread, but same context
         invoker.start();
      }
      
      return invoker;
   }
   
   /**
    * Provides a mechanism of performing asynchronous requests, by hiding the operating state of
    * this server (State-free or State-managed).
    */
   private abstract class AsyncInvoker
   {
      /** Flag indicating the server is still available for */
      private boolean terminated;
      
      /**
       * Invoke the async request, based on the operating mode.
       * 
       * @param    request
       *           The request to be invoked.
       */
      public abstract void invoke(AsyncRequestImpl request);
      
      /**
       * Start this async invoker.
       */
      public abstract void start();
      
      /**
       * Cancel all the active requests.
       * 
       * @return    <code>true</code> if cancellation was possible.
       */
      public abstract boolean cancelRequests();
      
      /**
       * Terminate this async invoker by setting the {@link #terminated} flag to <code>true</code>.
       */
      public synchronized void terminate()
      {
         this.terminated = true;
      }
      
      /**
       * Check if this async invoker is terminated.
       * 
       * @return   The state of the {@link #terminated} flag.
       */
      public synchronized boolean isTerminated()
      {
         return terminated;
      }
   }
   
   /**
    * Asynchronous invoker implementation for Session-managed operating mode.  In this mode, a
    * separate thread is started which listens for async requests added to the queue, and executes
    * them in a FIFO mode.
    */
   private class SessionManagedInvoker
   extends AsyncInvoker
   implements Runnable
   {
      /** Queue of async requests pending execution. */
      private ArrayDeque<AsyncRequestImpl> queue = new ArrayDeque<AsyncRequestImpl>();
      
      /** The current async request being executed. */
      private AsyncRequestImpl request = null;

      /**
       * In Session-managed operating mode, invoking an async request means adding it to the 
       * {@link #queue async queue}; from there, a specialized thread will pick it up and execute
       * it.
       * 
       * @param    request
       *           The async request to be executed.
       */
      @Override
      public void invoke(AsyncRequestImpl request)
      {
         synchronized (queue)
         {
            queue.addLast(request);
            queue.notify();
         }
      }

      /**
       * In Session-managed operating mode, a dedicated thread is started which listens for
       * async requests added to the {@link #queue}.
       */
      @Override
      public void start()
      {
         Thread t = new AssociatedThread(this);

         // register this thread as async
         SessionManager.get().registerAsyncThread(t);
         
         String serverName = helper.getServerName();
         String asyncName = String.format("Session-managed invoker [%s]", serverName);
         String name = LogHelper.createThreadName(asyncName);
         
         t.setName(name);
         t.start();
      }
      
      /**
       * Listens for incoming async requests and executes them in FIFO order.
       */
      @Override
      public void run()
      {
         try
         {
            // listen for requests while the server is connected
            while (!isTerminated() && _connected())
            {
               synchronized (queue)
               {
                  if (queue.isEmpty())
                  {
                     try
                     {
                        // as we can't sync on this object, make this wait a max amount of time
                        queue.wait(5000);
                     }
                     catch (InterruptedException e)
                     {
                        // restart the loop
                        continue;
                     }
                  }

                  if (queue.isEmpty())
                  {
                     // if we are still empty, restart the loop, maybe it was terminated
                     continue;
                  }
   
                  // get all requests and process them separately, in a FIFO order
                  request = queue.removeFirst();
               }
               
               request.execute(true);
            }
         }
         finally
         {
            // de-register this thread from the async list
            SessionManager.get().deregisterAsyncThread(Thread.currentThread());
         }
      }
      
      /**
       * Cancel all queued or running requests.
       */
      public boolean cancelRequests()
      {
         // process the queue
         synchronized (queue)
         {
            if (request != null && !request._isComplete())
            {
               // send STOP to current request and wait to complete
               request.stop();
            }

            // notify all requests that they are cancelled
            for (AsyncRequestImpl req : queue)
            {
               req.cancelled();
            }

            // clear the queue
            queue.clear();
         }
         
         return true;
      }
   }

   /**
    * Asynchronous invoker implementation for Session-free operating mode.  In this mode, each
    * async request is executed in its separate thread.
    */
   private class SessionFreeInvoker
   extends AsyncInvoker
   {
      /** A set of running requests. */
      private Set<AsyncRequestImpl> requests = new HashSet<AsyncRequestImpl>();

      /**
       * Invoking an async request in State-free mode needs to request to be executed in its
       * separated thread. For this, all running requests are registered in the {@link #requests}
       * set, to be able to cancel them.
       * 
       * @param    request
       *           The request to be executed.
       */
      @Override
      public void invoke(final AsyncRequestImpl request)
      {
         synchronized (requests)
         {
            requests.add(request);
         }

         // each request has its own thread
         Thread t = new AssociatedThread(new Runnable()
         {
            public void run()
            {
               try
               {
                  request.execute(true);
               }
               finally
               {
                  synchronized (requests)
                  {
                     requests.remove(request);
                  }
   
                  // de-register this thread from the async list
                  SessionManager.get().deregisterAsyncThread(Thread.currentThread());
               }
            }
         });
         
         // register this thread as async
         SessionManager.get().registerAsyncThread(t);
         
         String pname = request.getProcedureName().toStringMessage();
         String serverName = helper.getServerName();

         String asyncName = String.format("Session-free invoker [%s:%s]", serverName, pname);
         String name = LogHelper.createThreadName(asyncName);
         
         t.setName(name);
         t.start();
      }

      /**
       * This is a no-op, as the State-free invoker doesn't need a dedicated thread.
       */
      @Override
      public void start()
      {
         // no-op
      }

      /**
       * Cancel all running requests by sending the STOP conditions to them.
       * 
       * @return   <code>true</code> if cancel was possible.
       */
      @Override
      public boolean cancelRequests()
      {
         synchronized (requests)
         {
            for (AsyncRequestImpl request : requests)
            {
               // send STOP to this request
               if (!request._isComplete())
               {
                  // send stop to current request and wait it to complete
                  request.stop();
                  
                  // mark as cancelled too
                  request.cancelled();
               }
            }
            
            requests.clear();
         }

         return true;
      }
   }
}
